// cVector.cpp

#import "cVector.hpp"

#import "cPlane.hpp"
#import "cMatrix.hpp"

#if !defined(_180_OVER_PI)
#define _180_OVER_PI 0.017454
#endif


cVector::cVector()
{
  Set(0, 0, 0, 0);
}


BOOLEAN  cVector::Equal(cVector v)
{
  return ((mV[X]==v.mV[X]) && (mV[Y]==v.mV[Y]) && (mV[Z]==v.mV[Z]));
}


BOOLEAN  cVector::Unequal(cVector v)
{
  return ((mV[X]!=v.mV[X]) || (mV[Y]!=v.mV[Y]) || (mV[Z]!=v.mV[Z]));
}


void     cVector::Set(SCALAR x, SCALAR y, SCALAR z, SCALAR w)
{
  setVector(mV, x, y, z);
  mV[W] = w;
}


void     cVector::Set(SCALAR x, SCALAR y, SCALAR z)
{
  setVector(mV, x, y, z);
}


void     cVector::Set(cVector p0, cVector p1)
{
  vectorsub(mV, p0.mV, p1.mV);
}


INTEG      cVector::Perspective(SCALAR xzoom, SCALAR yzoom)
{
  if (mV[Z]<=0) return 1;
  mV[X] = xzoom * mV[X] / mV[Z];
  mV[Y] = yzoom * mV[Y] / mV[Z];
  mV[Z] = 1.0 / mV[Z];
  return 0;
}


void     cVector::Polar()
{
  cVector v;
  v.mV[X] = XYangle();
  v.mV[Y] = XZangle();
  v.mV[Z] = Len();
  *this = v;
}


void     cVector::Normalise()
{
  SCALAR vector_length = magnitude(mV);
  scalardiv(mV, mV, vector_length);
}


SCALAR   cVector::Len()
{
  return (magnitude(mV));
}


SCALAR   cVector::Dot(cVector v)
{
  return (dotprod(mV, v.mV));
}


void     cVector::Crs(cVector v)
{
  cross(mV, mV, v.mV);
}


void     cVector::Add(cVector v)
{
  vectoradd(mV, mV, v.mV);
}


void     cVector::Sub(cVector v)
{
  vectorsub(mV, mV, v.mV);
}


void     cVector::Mul(SCALAR s)
{
  scalarmul(mV, mV, s);
}


void     cVector::Div(SCALAR s)
{
  scalardiv(mV, mV, s);
}


void     cVector::Neg()
{
  negVector(mV);
}


ANGLE    cVector::XYangle()
{
//  return (atan2(mV[Y], mV[X]) * _180_OVER_PI);
  SCALAR s    = _180_OVER_PI * atan(xabs(mV[Y] / mV[X]));
  if ( mV[X] >= 0 ) {
    if ( mV[Y] >= 0 ) {
      return 90-(s);
    } else {
      return 90+(s);
    };
  } else {
    if ( mV[Y] < 0 ) {
      return 180+90-(s);
    } else {
     return 270+(s);
   };
 };
}


ANGLE    cVector::XZangle()
{
//  return (180 + atan2(mV[Z], mV[X]) * _180_OVER_PI);
  SCALAR s    = _180_OVER_PI * atan(xabs(mV[Z] / mV[X]));
  if ( mV[X] >= 0 ) {
    if ( mV[Z] >= 0 ) {
      return 90-(s);
    } else {
      return 90+(s);
    };
  } else {
    if ( mV[Z] < 0 ) {
      return 180+90-(s);
    } else {
      return 270+(s);
    };
  };
}


ANGLE    cVector::YZangle()
{
//  return (atan2(mV[Z], mV[Y]) * _180_OVER_PI);
  SCALAR s    = _180_OVER_PI * atan(xabs(mV[Z]) / mV[Y]);
  if ( mV[Y] >= 0 ) {
    if ( mV[Z] >= 0 ) {
      return 90-(s);
    } else {
      return 90+(s);
    };
  } else {
    if ( mV[Z] < 0 ) {
      return 180+90-(s);
    } else {
      return 270+(s);
    };
  };
}


INTEG cVector::FindInList(cVector *vList, INTEG n)
{
  INTEG i;
  for (i=0; i<n; i++) if (this->Equal(vList[i])) return i;
  return -1;
}


BOOLEAN cVector::InPolygon(INTEG n, cVector *vtx)
{
  INTEG i;
  cPlane edge;
  cVector mainNrm, nrm;
  makeNormal(mainNrm.mV, vtx[0].mV, vtx[1].mV, vtx[2].mV);
  mainNrm.Normalise();
  for (i=0; i<n; i++)
  {
    nrm = vtx[(i+1)%n];
    nrm.Sub(vtx[i]);
    nrm.Dot(mainNrm);
    edge.Set(nrm, vtx[i]);
    if (edge.Distance(*this) < 0) return 0;
  };
  return 1;
}


cVector cVector::GetUV(INTEG n, cVector *vtx, cVector *tex) // n textured points
{
  cMatrix   m;
  cPlane    plane;
  cVector   nrm;
  cVector   vcut[3], tcut[3];
  INTEG       i, a, b, c = 0;
  double    de, *d;

  d = new double[n];

  makeNormal(nrm.mV, vtx[0].mV, vtx[1].mV, vtx[2].mV);
  nrm.Normalise();
  m.RotX(90.0);
  m.Apply(nrm);
  plane.Set(nrm, *this);

  for (i=0; i<n; i++)
  {
    d[i] = plane.Distance(vtx[i]);
  };

  for (i=0; i<n; i++)
  {
    a = i;
    b = (i+1)%n;
    if (xsgn(d[a]) != xsgn(d[b])) {
      de = d[a] / (d[a] - d[b]);
      vcut[c].mV[X] = clif(vtx[a].mV[X], vtx[b].mV[X], de);
      vcut[c].mV[Y] = clif(vtx[a].mV[Y], vtx[b].mV[Y], de);
      vcut[c].mV[Z] = clif(vtx[a].mV[Z], vtx[b].mV[Z], de);
      tcut[c].mV[X] = clif(tex[a].mV[X], tex[b].mV[X], de);
      tcut[c].mV[Y] = clif(tex[a].mV[Y], tex[b].mV[Y], de);
      c++;
    };
    if (c >= 2) break;
  };

  a = 0;
  b = 1;
  c = 2;
  de = lineDistance(vcut[a].mV, this->mV) / lineDistance(vcut[a].mV, vcut[b].mV);
  vcut[c].mV[X] = clif(vcut[a].mV[X], vcut[b].mV[X], de);
  vcut[c].mV[Y] = clif(vcut[a].mV[Y], vcut[b].mV[Y], de);
  vcut[c].mV[Z] = clif(vcut[a].mV[Z], vcut[b].mV[Z], de);
  tcut[c].mV[X] = clif(tcut[a].mV[X], tcut[b].mV[X], de);
  tcut[c].mV[Y] = clif(tcut[a].mV[Y], tcut[b].mV[Y], de);

  delete d;

  return tcut[c];
}


void cVector::CenterOf(cVector *vList, INTEG n)
{
  INTEG i;
  Set(0, 0, 0);
  for (i=0; i<n; i++) Add(vList[i]);
  Div(n);
}


INTEG cVector::Nearest(cVector *vList, INTEG n)
{
  INTEG i, e = 0;
  SCALAR t, d = lineDistance(mV, vList[e].mV);
  for (i=1; i<n; i++)
  {
    t = lineDistance(mV, vList[i].mV);
    if (t < d) {
      d = t;
      e = i;
    };
  };
  return e;
}


INTEG cVector::Furthest(cVector *vList, INTEG n)
{
  INTEG i, e = 0;
  SCALAR t, d = lineDistance(mV, vList[e].mV);
  for (i=1; i<n; i++)
  {
    t = lineDistance(mV, vList[i].mV);
    if (t > d) {
      d = t;
      e = i;
    };
  };
  return e;
}


void cVector::Print(FILE *f)
{
  fprintf(f, "( %f, %f, %f )\n", mV[X], mV[Y], mV[Z]);
}


