// cMech.cpp

#include "cMech.h"


cMech::cMech(float* pos, float* rot)
{
  mRoles |= (1UL << cObject::ENTITY) | (1UL << cObject::MECH);
  
  mController = new cController(this, true);
  
  if (pos != NULL) memcpy(mPos, pos, sizeof(float) * 3);
  if (rot != NULL) memcpy(mRot, rot, sizeof(float) * 3);
  memset(mTwr, 0, sizeof(float) * 3);

  mRotators = new float[MAX_MECH_PARTS][3];
  for (int i = 0; i < MAX_MECH_PARTS; i++)
  {
    memset(mRotators, 0, sizeof(float) * 3);
  };
  
  mCurrentWeapon = 0;
  mTimer = 0;
  mJetOn = 0;
  mSpeed = 0;
  
  mCameraState = 1;
  
  mModel = NULL; // Deprecated: See header.
}


void cMech::onSpawn()
{
  // Find Model-Object.
  mModel = new cConstructModel((char*)"data/objects/mech0.txt");

  // Register position and rotation vector for Model-Object.
  mModel->setRotator(this->mRot);
  mModel->setTranslator(this->mPos);

  // Register some more rotation vectors for each limb you want to use.
  mModel->setConstructRotator((char*)"CTorsor", mRotators[CTorsor]);
  mModel->setConstructRotator((char*)"LUpLeg", mRotators[LUpLeg]);
  mModel->setConstructRotator((char*)"LLoLeg", mRotators[LLoLeg]);
  mModel->setConstructRotator((char*)"RUpLeg", mRotators[RUpLeg]);
  mModel->setConstructRotator((char*)"RLoLeg", mRotators[RLoLeg]);
  
  this->mountWeapon((char*)"CTorsor", &explosion);
  this->mWeapons.remove(&explosion);
  
  printf("Mech spawned\n");
  if (mModel == NULL) printf("-> warning no model\n");
}


void cMech::multEyeMatrix()
{
    //glPushMatrix();
        //glLoadIdentity();
        if (abs(mCameraState) == 1) {
            // Normal View from cockpit nothing to change
        } else if (abs(mCameraState) == 2) {
            glTranslatef(0, 0, -5);
            glRotatef(15, 1, 0, 0);
        } else if (abs(mCameraState) == 3) {
            glTranslatef(0, 0, -100);
            glRotatef(90, 1, 0, 0);
        }
        glRotatef(180, 0, 1, 0);
        if (mModel) glMultMatrixf(mModel->getConstructInvMatrix((char*)"Head"));
        if (mModel) glMultMatrixf(mModel->getInvMatrix());
        //glGetFloatv(GL_MODELVIEW_MATRIX, mPilot);
    //glPopMatrix();
    //return mPilot;
}


void cMech::Pattern(float t, char *pattern)
{
  int l = strlen(pattern);
  int s = int(t/90.0) % l;
  char c = pattern[s];
  switch (c) {
    case 'l': ChassisLR(+5); break;
    case 'r': ChassisLR(-5); break;
    case 'n': break;
    default:  break;
  };
}


void cMech::ChassisLR(float r)
{
  const float limit = 1.75;
  float e = r;
  if (e>+limit) e = +limit;
  if (e<-limit) e = -limit;
  mRot[Y] += e;
  mRot[Y] = fmod(mRot[Y], 360);
  if (mRot[Y] < 0) mRot[Y] += 360;
}


void cMech::ChassisUD(float r)
{
  float e = r;
  if (e>+1) e = +1;
  if (e<-1) e = -1;
  mRot[X] += e;
  mRot[X] = fmod(mRot[X], 360);
  if (mRot[X] < 0) mRot[X] += 360;
}


void cMech::TowerLR(float r)
{
  float e = r;
  if (e>+2) e = +2;
  if (e<-2) e = -2;
  mTwr[Y] += e;
  if (mTwr[Y] > +75) mTwr[Y] = +75;
  if (mTwr[Y] < -75) mTwr[Y] = -75;
}


void cMech::TowerUD(float r)
{
  float e = r;
  if (e>+2) e = +2;
  if (e<-2) e = -2;
  mTwr[X] += e;
  if (mTwr[X] > +75) mTwr[X] = +75;
  if (mTwr[X] < -75) mTwr[X] = -75;
}


void cMech::fireAllWeapons(cObject *target)
{
  list<cWeapon *>::iterator it;
  cWeapon *weapon;
  for (it = mWeapons.begin(); it != mWeapons.end(); it++)
  {
    weapon = *it;
    weapon->fire(target);
  };
}


void cMech::fireWeapon(int n, cObject *target)
{
  list<cWeapon *>::iterator it;
  cWeapon *weapon;
  int i = 0;
  for (it = mWeapons.begin(); it != mWeapons.end(); it++)
  {
    if (i == n)
    {
      weapon = *it;
      weapon->fire(target);
      break;
    }
    i++;
  }
}


void cMech::mountWeapon(char* point, cWeapon *weapon)
{
    if (mModel == NULL) {
        printf("mountWeapon(): no model\n");
        return;
    }
    weapon->mBasefv = mModel->getMatrix();
    weapon->mMountfv = mModel->getConstructMatrix(point);
    assert(weapon->mMountfv != NULL);
    // The weapon isn't spawned but it needs to know the world
    // and its owner because the owner may be ignored on filtering targets.
    weapon->mWorld = mWorld;
    weapon->mOwner = this;
    mWeapons.push_back(weapon);
}


void cMech::Walk(float spf)
{
  float a = mSpeed / 100.0;
  float sp = a * spf * 5.0;
  float jt = mJetOn / 100.0 * spf * 15.0;
  const float e = 0.0174545;
  mTimer += spf;
  float t = mTimer * 180;
  a = a*0.75;

  float vec[3] = { 0, 0, 0 };

  memcpy(mRotators[CTorsor], mTwr, sizeof(float) * 3);

  if (mPos[Y] < 0.5) {
    memset(vec, 0, sizeof(float) * 3);
    vec[X] = (+sin(t*e)*30)*a;
    memcpy(mRotators[LUpLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = (-sin(t*e)*30)*a;
    memcpy(mRotators[RUpLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = (+sin(t*e*xsgn(a))*15+15)*fabs(a);
    memcpy(mRotators[LLoLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = (-sin(t*e*xsgn(a))*15+15)*fabs(a);
    memcpy(mRotators[RLoLeg], vec, sizeof(float) * 3);
  } else {
    memset(vec, 0, sizeof(float) * 3);
    vec[X] = -10;
    memcpy(mRotators[LUpLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = -10;
    memcpy(mRotators[RUpLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = +20;
    memcpy(mRotators[LLoLeg], vec, sizeof(float) * 3);

    memset(vec, 0, sizeof(float) * 3);
    vec[X] = +20;
    memcpy(mRotators[RLoLeg], vec, sizeof(float) * 3);
  };

  mPos[X] += sp*sin(mRot[Y]*e)*cos(mRot[X]*e);
  mPos[Y] += jt;
  mPos[Z] += sp*cos(mRot[Y]*e)*cos(mRot[X]*e);

  mPos[X] -= mWorld->mGravity[X] * spf;
  mPos[Y] -= mWorld->mGravity[Y] * spf;
  mPos[Z] -= mWorld->mGravity[Z] * spf;

  // if ground contact.
  if (mPos[Y] < 0.5) mPos[Y] += sin((t*2.0 - 10)*e)*sp/2;
  if (mPos[Y] < 0) mPos[Y] = 0;
}


void cMech::animateSelf(float spf)
{
    if (this->mHp <= 0 && explosion.ready() && mWorld != NULL) {
        this->ChassisLR(0);
        this->TowerLR(0);
        this->TowerUD(0);
        mWorld->fragObject(this);
        return;
    }
      
  if (mController && mController->mEnabled) {
      if (mPad) mPad->reset();
      mController->processSelf();
  }
    
    if (mHp <= 0) {
        mSpeed = 0;
    }
  
  if (mHp > 0 && mPad) {
      //mPad->print();
      
    this->TowerLR(50*spf * -mPad->getAxis(cPad::MECH_TURRET_LR_AXIS));
    this->TowerUD(50*spf * -mPad->getAxis(cPad::MECH_TURRET_UD_AXIS));

    this->ChassisLR(50*spf * -mPad->getAxis(cPad::MECH_CHASSIS_LR_AXIS));
    float tdelta = 50*spf * -mPad->getAxis(cPad::MECH_THROTTLE_AXIS);
    if (tdelta != 0.0f) {
      mSpeed += 10*tdelta;
      if (mSpeed > +220) mSpeed = +220; else
      if (mSpeed < -100) mSpeed = -100;
    }
    
    if (mPad->getButton(cPad::MECH_FIRE_BUTTON)) {
      if (true) {
        mCurrentWeapon %= mWeapons.size();
        fireWeapon(mCurrentWeapon, mTarget);
        mCurrentWeapon++;
        mCurrentWeapon %= mWeapons.size();
      } else {
        fireAllWeapons(mTarget);
      }
    }
    
    if (mPad->getButton(cPad::MECH_JET_BUTTON)) {
        mJetOn += (100-mJetOn) * 0.1;
    } else {
        mJetOn -= (mJetOn * 0.05);
    }
    
    if (mPad->getButton(cPad::MECH_CAMERA_BUTTON)) {
        // Only if Camera State is not transitional
        // then switch to next perspective on button press.
        if (mCameraState > 0) {
            mCameraState = (1 + (mCameraState % 3));
            mCameraState *= -1; // Set transitional.
        }
    } else {
        // Set Camera State as fixed.
        mCameraState = abs(mCameraState);
    }
  }
  
  Walk(spf);
  
  for (list<cWeapon *>::iterator it = mWeapons.begin(); it != mWeapons.end(); it++)
  {
    cWeapon* weapon = *it;
    weapon->animate(spf);
  }
  explosion.animate(spf);
}


void cMech::transformSelf()
{
  if (mModel) mModel->transformSelf();
  for (list<cWeapon *>::iterator it = mWeapons.begin(); it != mWeapons.end(); it++)
  {
    cWeapon* weapon = *it;
    weapon->transform();
  };
}


void cMech::drawSolid()
{
  if (mModel) mModel->drawSelf();
  for (list<cWeapon *>::iterator it = mWeapons.begin(); it != mWeapons.end(); it++)
  {
    cWeapon* weapon = *it;
    weapon->drawSolid();
  };
  explosion.drawSolid();
}


void cMech::drawEffect()
{
  for (list<cWeapon *>::iterator it = mWeapons.begin(); it != mWeapons.end(); it++)
  {
    cWeapon* weapon = *it;
    weapon->drawEffect();
  };
  explosion.drawEffect();
}


void cMech::drawHUD()
{
    glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT);
    glLineWidth(2);    
    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_FOG);

        glPushUnitOrthoProjection();

            glPushMatrix();
            glLoadIdentity();

                // ...
                glDisable(GL_DEPTH_TEST);
                glDisable(GL_TEXTURE_2D);
                if (!true) {
                    glBegin(GL_LINE_STRIP);
                        glColor4f(1,0.9,0,1);
                        glVertex3f(0.1, 0.1, 0);
                        glVertex3f(0.9, 0.1, 0);
                        glVertex3f(0.9, 0.9, 0);
                        glVertex3f(0.1, 0.9, 0);
                        glVertex3f(0.1, 0.1, 0);
                    glEnd();
                }
                
                float f = 200;
                
                glBegin(GL_LINES);
                    glColor4f(0,0.2,0.7,1);
                    glVertex3f(0.5, 0.1, 0);
                    glVertex3f(0.5 + this->mTwr[1]/f, 0.1, 0);
                    
                    glVertex3f(0.5 + this->mTwr[1]/f, 0.08, 0);
                    glVertex3f(0.5 + this->mTwr[1]/f, 0.12, 0);
                    
                    glVertex3f(0.5, 0.9, 0);
                    glVertex3f(0.5 + this->mTwr[1]/f, 0.9, 0);
                    
                    glVertex3f(0.5 + this->mTwr[1]/f, 0.88, 0);
                    glVertex3f(0.5 + this->mTwr[1]/f, 0.92, 0);
                glEnd();
                
                glBegin(GL_LINES);
                    glColor4f(0,0.2,0.7,1);
                    glVertex3f(0.1, 0.5, 0);
                    glVertex3f(0.1, 0.5 + this->mTwr[0]/f, 0);
                    
                    glVertex3f(0.08, 0.5 + this->mTwr[0]/f, 0);
                    glVertex3f(0.12, 0.5 + this->mTwr[0]/f, 0);
                    
                    glVertex3f(0.9, 0.5, 0);
                    glVertex3f(0.9, 0.5 + this->mTwr[0]/f, 0);
                    
                    glVertex3f(0.88, 0.5 + this->mTwr[0]/f, 0);
                    glVertex3f(0.92, 0.5 + this->mTwr[0]/f, 0);
                glEnd();

            glPopMatrix();

        glPopProjection();
    glPopAttrib();
}


cObject::rImpact* cMech::damageByCylinder(float* a, float* b, float radius, float damage, bool all)
{
    if (mModel == NULL) return NULL;
    
    cConstructModel::rImpact* result = NULL;
    if (mModel) mModel->collideLine(a, b, result);
    if (result == NULL) {
        //printf("no collision, result == NULL\n");
        return NULL;
    }
    
    cObject::rImpact* im = new cObject::rImpact;
    memcpy(im->pos, result->intsec_point, sizeof(float)*3);
    memcpy(im->nrm, result->intsec_normal, sizeof(float)*3);
    delete result;
    
    printf("damage\n");
    mHp -= damage;
    
    //if (mHp <= 0) mWorld->fragObject(this);
    if (mHp <= 0) explosion.fire(NULL);
    
    return im;
}


cObject* cMech::enemyNearby()
{
  if (mWorld == NULL) return NULL;
  cObject* result = NULL;
  // Find all objects belonging to any enemy party/role.
  list<cObject*>* roles = mWorld->filterByRole(this, this->mEnemies, false, -1);
  // Filter enemies by distance, choose any single one in diameter.
  list<cObject*>* scan = mWorld->filterByRange(this, this->mPos, 0.0f, 50.0f, 1, roles);
  if (!scan->empty()) result = scan->front();
  delete scan;
  delete roles;
  return result;
}


float cMech::inMeeleRange()
{
  if (mTarget == NULL) return 0.0f;
  float a = 16;
  float b = 24;
  float d = lineDistance(mPos, mTarget->mPos);
  if (d < a) return 1.0f;
  if (d > b) return 0.0f;
  return (1.0f - (d - a) / (b - a));
}


float cMech::inWeaponRange()
{
  // TODO should depend on ready to fire weapons properties.
  if (mTarget == NULL) return 0.0f;
  float a = 36+10;
  float b = 44+10;
  float d = lineDistance(mPos, mTarget->mPos);
  if (d < a) return 1.0f;
  if (d > b) return 0.0f;
  return (1.0f - (d - a) / (b - a));
}


float cMech::inTargetRange()
{
  if (mTarget == NULL) return 0.0f;
  float a = 56;
  float b = 64;
  float d = lineDistance(mPos, mTarget->mPos);
  if (d < a) return 1.0f;
  if (d > b) return 0.0f;
  return (1.0f - (d - a) / (b - a));
}


void cMech::do_moveTowards()
{
  if (mPad == NULL) return;
  
  float dir[3] = { this->mPos[0] - mTarget->mPos[0], this->mPos[1] - mTarget->mPos[1], this->mPos[2] - mTarget->mPos[2] };
  float  rd = 0.0f;
  rd = (180 + atan2(dir[X], dir[Z]) * 180.0f/3.145479f) - this->mRot[Y];
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.
  mPad->setAxis(cPad::MECH_CHASSIS_LR_AXIS, -rd);
  
  float thr = 20.0f * (1.0f - fabs(rd) / 360.0f);
  //printf("Throttle: %.2f\n", thr);
  mPad->setAxis(cPad::MECH_THROTTLE_AXIS, (char) -thr);
}


void cMech::do_moveNear()
{
  if (mPad == NULL) return;

  float dir[3] = { this->mPos[0] - mTarget->mPos[0], this->mPos[1] - mTarget->mPos[1], this->mPos[2] - mTarget->mPos[2] };
  float  rd = 0.0f;
  rd = (180 + atan2(dir[X], dir[Z]) * 180.0f/3.145479f) - this->mRot[Y];
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.
  mPad->setAxis(cPad::MECH_CHASSIS_LR_AXIS, -rd);
  
  float d = lineDistance(this->mPos, mTarget->mPos);
  float f = (d - 11);
  if (f < -5) f = -5;
  if (f > +10) f = +10;
  mPad->setAxis(cPad::MECH_THROTTLE_AXIS, -f*0.2f);
}


void cMech::do_aimAt()
{
  if (mPad == NULL) return;

  float dir[3] = { this->mPos[0] - mTarget->mPos[0], this->mPos[1] - mTarget->mPos[1], this->mPos[2] - mTarget->mPos[2] };
  float rd = (180 + atan2(dir[X], dir[Z]) * 180.0f/3.145479f) - (this->mRot[Y] + this->mTwr[Y]);
  if (fabs(rd) > 180) rd = -(rd / fabs(rd)) * (360 - fabs(rd)); // -sign(rd) * circle other way around.
  mPad->setAxis(cPad::MECH_TURRET_LR_AXIS, -rd/2.0f);
}


void cMech::do_fireAt()
{
  if (mPad == NULL) return;
  if (mTarget == NULL) return;

  float dir[3] = { this->mPos[0] - mTarget->mPos[0], this->mPos[1] - mTarget->mPos[1], this->mPos[2] - mTarget->mPos[2] };
  float  rd = (180 + atan2(dir[X], dir[Z]) * 180.0 / M_PI) - this->mRot[Y];
  if (rd > 180) rd = 360 - rd;
  //printf("Target Angle: %.2f   Own Angle: %.2f    Angle Diff: %.2f\n", (180 + atan2(dir[X], dir[Z]) * 180.0 / M_PI), this->mRot[Y], rd);
  
  if (rand()%6 == 0) {
    if (fabs(rd) < 3) {
      mPad->setButton(cPad::MECH_FIRE_BUTTON, true);
      //printf("fire\n");
    };
  };
}


void cMech::do_idle()
{
  if (mPad == NULL) return;
  mPad->reset();
  //mPad->setAxis(cPad::MECH_TURRET_UD_AXIS, -1.0f);
}


void cMech::do_aimFor(cObject* target)
{
  mTarget = target;
}

