/**
 * File:   cWorld.h
 * Author: hackCraft
 *
 * Created on March 28, 2008, 10:23 PM (1999)
 */

#ifndef _CWORLD_H
#define _CWORLD_H

class cWorld;
#include "psi3d/math3d.h"

#include <list>

using namespace std;

class cBackground;
class cObject;
class cMission;


class cWorld {
public:
  static cWorld* sWorld;
  cBackground* mBackground; // Background Object
  cMission* mMission; // Mission Setup and Objective-Controller.
  float mGravity[3]; // Downward force vector (or 0,0,0 in space).

  list<cObject*> mObjects; // Contains alive (not yet fragged) objects.
  list<cObject*> mCorpses; // Contains fragged (but not yet deleted) objects.

  //map<string, void *> mGamevars;
  //map<string, string> mGamevars;

  float mTime; // Simulation time in 24 Hours + fractions of hours.
  float mSPF; // seconds per frame (1.0f/fps).
  float mFPS; // frames per second (1.0f/spf).

public: // Constructor, Object-Management and Drawing:

  cWorld();
  
public: // Spawning And Fragging

  // Spawns object imediatly.
  void spawnObject(cObject *object);
  
  // Sets object fragged and inserts it into the mBufferedCorpses-List
  // so it can be deleted later on by bagFragged().
  void fragObject(cObject *object);
  
  // On frag Objects aren't deleted instantly.
  // Instead Objects reside inside a Graveyard-list
  // and are finally removed and deleted by this call.
  // Possibly called very rarely within a Mission at transitional moments.
  void bagFragged();

public: // Simulation Step.

  // Call every frame to update (animation) timers.
  void advanceTime(int deltamsec);
  void animateObjects();
  void transformObjects();
  void drawSolid();
  void drawEffect();

public: // World-Filtering, World-Scanning, World-Sense for objects.
  
  // Debug function to print out an object-list.
  // With a parameter of NULL/no parameter at all prints all objects within the world.
  // Warning: Printing all objects could lead to massive load on the console. 
  void printNames(list<cObject*>* objects = NULL);

  // Filters the given objectlist (or all mObjects if NULL) by roles (ORed Bitmask) (excluding ex Object).
  // There is the option to match all roles (true=AND) or to select any given role (false=OR).
  // The returned list is fresh allocated - caller has to delete it.
  list<cObject*>* filterByRole(cObject* ex, unsigned long rolemask, bool all, int maxamount, list<cObject*>* objects = NULL);
  
  // Returns a List of objects which are within minimum and maximum range (excluding ex Object).
  // The returned list is fresh allocated - caller has to delete it.
  list<cObject*>* filterByRange(cObject* ex, float* origin, float minrange, float maxrange, int maxamount, list<cObject*>* objects = NULL);
  
  // Returns a List of objects which have the specified name.
  // The returned list is fresh allocated - caller has to delete it.
  list<cObject*>* filterByName(cObject* ex, char* name, int maxamount, list<cObject*>* objectlist = NULL);
  
  // Returns a List of objects whose bounding volume (approximately) intersects with the cylinder
  // described by the volume of certain diameter around a line.
  // The returned list is fresh allocated - caller has to delete it.
  list<cObject*>* filterByCylinder(cObject* ex, float* pointa, float* pointb, float radius, int maxamount, list<cObject*>* objects = NULL);
};

#endif
